!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module util
contains
  real(kind=8) function cellvol(aa)
    !
    !       calculates the cell volume of aa
    !
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    !
    real(kind=8) v
    !
    v=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))
    cellvol=abs(v)
    return
  end function cellvol
  !
  subroutine outer_product(a,b,c)
    !
    !       calculates the outer product c = a \cross b
    !
    implicit none
    real(kind=8),intent(in):: a(3),b(3)
    real(kind=8),intent(out):: c(3)
    !        
    c(1) = a(2)*b(3)-a(3)*b(2)
    c(2) = a(3)*b(1)-a(1)*b(3)
    c(3) = a(1)*b(2)-a(2)*b(1)
    return
  end subroutine outer_product
  !
  integer function suplatp(aa,r)
    implicit none
    real(kind=8),intent(in)::aa(3,3),r
    !
    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    real(kind=8):: x(3),d
    !
    ! calculate cell diameter d.
    !
    x(:) = aa(:,1) + aa(:,2) + aa(:,3)
    d = sqrt(dot_product(x,x))
    !
    x(:) = aa(:,1) + aa(:,2) - aa(:,3)
    d = max(d, sqrt(dot_product(x,x)))
    !
    x(:) = aa(:,1) - aa(:,2) + aa(:,3)
    d = max(d, sqrt(dot_product(x,x)))
    !
    x(:) = aa(:,1) - aa(:,2) - aa(:,3)
    d = max(d, sqrt(dot_product(x,x)))
    !
    suplatp = int((4.0d0*pi/3.0d0)*(r+d)**3/cellvol(aa))
    return
  end function suplatp
  !
  subroutine LatticeInterval(aa,h)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8),intent(out):: h(3)
    !
    real(kind=8):: hh(3,3),htmp
    !
    ! hh(*,1) = aa(*,2) \times aa(*,3)
    ! hh(*,2) = aa(*,3) \times aa(*,1)
    ! hh(*,3) = aa(*,1) \times aa(*,2)
    !
    call outer_product(aa(:,2),aa(:,3),hh(:,1))
    call outer_product(aa(:,3),aa(:,1),hh(:,2))
    call outer_product(aa(:,1),aa(:,2),hh(:,3))
    !
    ! normalize hh(*,1),hh(*,2),hh(*,3)
    !
    htmp=sqrt(dot_product(hh(:,1),hh(:,1)))
    hh(:,1)=hh(:,1)/htmp
    htmp=sqrt(dot_product(hh(:,2),hh(:,2)))
    hh(:,2)=hh(:,2)/htmp
    htmp=sqrt(dot_product(hh(:,3),hh(:,3)))
    hh(:,3)=hh(:,3)/htmp
    !
    ! h(1): interval of the plane (1,0,0) etc.
    !
    h(1)=abs(dot_product(aa(:,1),hh(:,1)))
    h(2)=abs(dot_product(aa(:,2),hh(:,2)))
    h(3)=abs(dot_product(aa(:,3),hh(:,3)))
    return
  end subroutine LatticeInterval
  !
  subroutine fctck(n,m2,m3,m5,info)
    implicit none
    integer,intent(in):: n
    integer,intent(out):: m2, m3, m5, info
    !
    integer:: i
    !
    i=n
    m2 = 0
    m3 = 0
    m5 = 0
    info = 0
    !
    do while (i .ne. 1)
       if (mod(i,2) .eq. 0) then
          m2 = m2 + 1
          i = i / 2
       else if (mod(i,3) .eq. 0) then
          m3 = m3 + 1
          i = i / 3
       else if (mod(i,5) .eq. 0) then
          m5 = m5 + 1
          i = i / 5
       else
          info = 1
          exit
       end if
    end do
    return
  end subroutine fctck
!
  integer function algn235(inr,gctr)
    implicit none
    integer,intent(in):: inr,gctr
!
    integer:: nr,m2,m3,m5,info
    integer:: t2, t3, t5
!
    call fctck(gctr,t2,t3,t5,info)
    if (info.eq.1) then
       write(6,*) 'algn235: gctr is not a multiple of 2,3,5'
       stop
    end if
    nr = inr
    call fctck(nr,m2,m3,m5,info)
    do while ( (info .eq. 1) .or. (m2.lt.t2) &
         .or. (m3.lt.t3) .or. (m5.lt.t5) )
       nr = nr + 1
       call fctck(nr,m2,m3,m5,info)
    end do
    algn235 = nr
    return
  end function algn235
!
  logical function normconsp(nl,nr,maxr,ifcr)
    implicit none
    integer,intent(in):: nl, nr(nl), maxr, ifcr(maxr,nl)
    integer:: il, ir
    logical:: cnd

    cnd = .true.
!
! These loops are very short.
!
!POPTION NOPARALLEL
    do il=1,nl
!POPTION NOPARALLEL
      do ir=1,nr(il)
         if (ifcr(ir,il) .gt. 0) then
            cnd = .false.
         end if
      end do
    end do
    normconsp = cnd
    return
  end function normconsp

  logical function hasinvsym(ng,nnp,rg,pg)
    implicit none
    integer,intent(in):: ng,nnp,rg(3,3,ng), pg(3,ng)

    integer:: ig
    logical:: retval

    retval = .false.
    do ig=1,ng
       if( (rg(1,1,ig).eq.-1) .and. (rg(2,1,ig).eq.0) &
                              .and. (rg(3,1,ig).eq.0) .and. &
           (rg(1,2,ig).eq.0) .and. (rg(2,2,ig).eq.-1) &
                             .and. (rg(3,2,ig).eq.0) .and. &
           (rg(1,3,ig).eq.0) .and. (rg(2,3,ig).eq.0) &
                             .and. (rg(3,3,ig).eq.-1) .and. &
           (mod(pg(1,ig),nnp).eq.0) .and. &
           (mod(pg(2,ig),nnp).eq.0) .and. &
           (mod(pg(3,ig),nnp).eq.0) ) then
          retval = .true.
       end if
    end do
    hasinvsym = retval
    return
  end function hasinvsym

  subroutine mkstrfac(nsi,asi,nkm,kg,strfac)
    implicit none
    integer,intent(in):: nsi, nkm, kg(3,nkm)
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(out):: strfac(nkm,2,nsi)

    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    real(kind=8):: ph
    integer:: it,ig

    do it=1,nsi
       do ig=1,nkm
          ph=2.0d0*pi*(asi(1,it)*kg(1,ig) &
               +asi(2,it)*kg(2,ig)+asi(3,it)*kg(3,ig))
          strfac(ig,1,it) = cos(ph)
          strfac(ig,2,it) = -sin(ph)
       end do
    end do
    return
  end subroutine mkstrfac

  real(kind=8) function difflpt(nvxyz, old, new)
    implicit none
    integer,intent(in):: nvxyz
    complex(kind=8),intent(in):: old(nvxyz), new(nvxyz)

    integer:: i
    complex(kind=8):: diff
    real(kind=8):: r2
      
    r2 = 0.0d0
    do i=1,nvxyz
       diff =  new(i) - old(i)
       r2 = r2 + dble(diff)*dble(diff) + aimag(diff)*aimag(diff)
    end do
!
!   \frac{1}{\Omega} \int_{\Omega} |V_{new}(r) - V_{old}(r)|^2 dr
! = \sum_{G} ( \frac{1}{\Omega} |V_{new}(G) - V_{old}(G)| )^2
!
    difflpt = r2

    return
  end function difflpt

  !  
  ! from Research Institute for Mathematical Sciences, Kyoto University
  ! Ko^kyu^roku vol. 1032, 21-45 (1998)
  ! Satoru Tanaka, Shuya Kanagawa
  ! pseudo randoum number B using Schrage technique
  !
  real(kind=8) function ran0(state)
    implicit none
    integer:: state

    integer,parameter:: a = 16807 ! = 7^5
    integer,parameter:: m = 2147483647 ! = 2^31 - 1
    integer,parameter:: r = 2836 ! = m/a
    integer,parameter:: q = 127773 ! = mod(m,a)
    integer:: tmp

    if (state.eq.0) then
       write(6,*) 'ran0: state = 0'
       stop
    end if

    tmp = state/q
    state = a*(state - q*tmp) - r*(state/q)
    if (state.lt.0) then
       state = state + m
    end if

    ran0 = dble(1.0d0)/dble(m) * state
    return
  end function ran0

  ! generate gaussian distribution G, from uniform distribution U
  ! E[U] = 1/2, V[U] = 1/12
  ! use  central limit theorem.
  ! ( 1/N \sum U_i - E[U] ) / ( V[U] / \sqrt N ) -> N(0,1)
  ! G = N(0,1)*sd = N(0,sd)
  real(kind=8) function rndgauss(sd,idum)
    implicit none
    real(kind=8),intent(in):: sd
    integer:: idum
    integer,parameter:: rnk = 12
    integer:: i
    real(kind=8):: g

    g = 0.0d0
    do i=1,rnk
       g = g + ran0(idum)
    end do
    g = g/rnk
    g = sqrt(dble(rnk))*(g - 0.5d0)*sqrt(12.0d0)
    rndgauss = g*sd
    return
  end function rndgauss

  integer function grtcomdiv(a,b)
    implicit none
    integer,intent(in):: a,b
    integer:: p,q,r

    if (a .eq. 0) then
       grtcomdiv = b
       return
    else if (b .eq. 0) then
       grtcomdiv = a
       return
    end if

    if (a .ge. b) then
       p = a
       q = b
    else
       p = b
       q = a
    end if

    do while (.true.)
       r = mod(p,q)
       if (r.eq.0) exit
       p = q
       q = r
    end do
    grtcomdiv = q
    return
  end function grtcomdiv

  subroutine grtcomtrns(ng,pg,nnp,gctr)
    implicit none
    integer,intent(in):: ng, pg(3,ng), nnp
    integer,intent(out):: gctr(3)

    integer:: i,ig
    do i=1,3
       gctr(i) = nnp
       do ig=1,ng
          gctr(i) = grtcomdiv(gctr(i),modulo(pg(i,ig),nnp))
       end do
       gctr(i) = nnp/gctr(i)
    end do
    return
  end subroutine grtcomtrns

!*********************************************************************
!  ACCORDING TO SLATER DEFINITION
!                SEE "The Calculation of Molecular Orbitals"
!  YLM(NG2Q,I) FOR YLM_L^M
!     1:(0, 0),2:(1,-1),3:(1, 0),4:(1, 1)
!     5:(2,-2),6:(2,-1),7:(2, 0),8:(2, 1),9:(2,2)
!          1992-07-17  Jun Yamauchi
!********************************************************************
  subroutine getylb(nkm,sk,kg,bb,ylm,lmax)
    implicit none
    integer,intent(in):: nkm,kg(3,nkm),lmax
    real(kind=8),intent(in):: bb(3,3),sk(3)
    real(kind=8):: ylm(nkm,lmax**2)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    real(kind=8):: s1,s2,s3,s0,f00,f10,f20,f21,f22,r,r2
    integer:: lmax2,ista,il,ig

    s1=bb(1,1)*(sk(1)+kg(1,1))+bb(1,2)*(sk(2)+kg(2,1)) &
         +bb(1,3)*(sk(3)+kg(3,1))
    s2=bb(2,1)*(sk(1)+kg(1,1))+bb(2,2)*(sk(2)+kg(2,1)) &
         +bb(2,3)*(sk(3)+kg(3,1))
    s3=bb(3,1)*(sk(1)+kg(1,1))+bb(3,2)*(sk(2)+kg(2,1)) &
         +bb(3,3)*(sk(3)+kg(3,1))
    s0=s1*s1+s2*s2+s3*s3
    lmax2=lmax*lmax

    f00=sqrt( 1.d0 / ( 4.d0*pi) )
    f10=sqrt( 3.d0 / ( 4.d0*pi) )
    f20=sqrt( 15.d0/pi )*0.5d0
    f21=sqrt( 5.0d0/pi )*0.25d0
    f22=f20*0.5d0
    if(s0.eq.0.d0) then
       ylm(1,1)= f00
       do il=2,lmax2
          ylm(1,il)=0.d0
       end do
       ista=2
    else
       ista=1
    endif
    if (lmax.eq.1) then
       do ig=ista,nkm
          ylm(ig,1)= f00
       end do
    else if (lmax.eq.2) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          ylm(ig,1)= f00
          ylm(ig,2)=-f10*s2/r
          ylm(ig,3)= f10*s3/r
          ylm(ig,4)=-f10*s1/r
       end do
    else if (lmax.eq.3) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          r2=r*r
          ylm(ig,1)= f00
          ylm(ig,2)=-f10*s2/r
          ylm(ig,3)= f10*s3/r
          ylm(ig,4)=-f10*s1/r
          ylm(ig,5)= f20*s1*s2/r2
          ylm(ig,6)=-f20*s2*s3/r2
          ylm(ig,7)= f21*(3.0d0*s3**2/r2-1.0d0)
          ylm(ig,8)=-f20*s3*s1/r2
          ylm(ig,9)= f22*(s1**2-s2**2)/r2
       end do
    else
       write(6,*)'LMAX must be >= 3.  LMAX=',lmax
       stop 'GETYL'
    end if
    return
  end subroutine getylb

  subroutine getyls(nkm,sk,kg,bb,ylms,lmax)
    implicit none
    integer,intent(in):: nkm,kg(3,nkm),lmax
    real(kind=8),intent(in):: bb(3,3),sk(3)
    real(kind=8):: ylms(nkm,3,lmax**2)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    real(kind=8):: s1,s2,s3,s0,f10,f20,f21,f22,r
    integer:: ista,ig

    s1=bb(1,1)*(sk(1)+kg(1,1))+bb(1,2)*(sk(2)+kg(2,1)) &
         +bb(1,3)*(sk(3)+kg(3,1))
    s2=bb(2,1)*(sk(1)+kg(1,1))+bb(2,2)*(sk(2)+kg(2,1)) &
         +bb(2,3)*(sk(3)+kg(3,1))
    s3=bb(3,1)*(sk(1)+kg(1,1))+bb(3,2)*(sk(2)+kg(2,1)) &
         +bb(3,3)*(sk(3)+kg(3,1))
    s0=s1*s1+s2*s2+s3*s3

    f10=sqrt( 3.d0 / ( 4.d0*pi) )
    f20=sqrt( 15.d0/pi )*0.5d0
    f21=sqrt( 5.0d0/pi )*0.25d0
    f22=f20*0.5d0
    if(s0.eq.0.d0) then
       ylms(1,:,:) = 0.0d0
       ista=2
    else
       ista=1
    endif
    if (lmax.eq.1) then
       do ig=ista,nkm
          ylms(ig,:,1)= 0.0d0
       end do
    else if (lmax.eq.2) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          ylms(ig,:,1)= 0.0d0
          ylms(ig,1,2)= 0.0d0
          ylms(ig,2,2)= -f10
          ylms(ig,3,2)= 0.0d0
          ylms(ig,1,3)= 0.0d0
          ylms(ig,2,3)= 0.0d0
          ylms(ig,3,3)= f10
          ylms(ig,1,4)= -f10
          ylms(ig,2,4)= 0.0d0
          ylms(ig,3,4)= 0.0d0
       end do
    else if (lmax.eq.3) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          ylms(ig,:,1)= 0.0d0
          ylms(ig,1,2)= 0.0d0
          ylms(ig,2,2)= -f10
          ylms(ig,3,2)= 0.0d0
          ylms(ig,1,3)= 0.0d0
          ylms(ig,2,3)= 0.0d0
          ylms(ig,3,3)= f10
          ylms(ig,1,4)= -f10
          ylms(ig,2,4)= 0.0d0
          ylms(ig,3,4)= 0.0d0
          ylms(ig,1,5)= f20*s2/r
          ylms(ig,2,5)= f20*s1/r
          ylms(ig,3,5)= 0.0d0
          ylms(ig,1,6)= 0.0d0
          ylms(ig,2,6)= -f20*s3/r
          ylms(ig,3,6)= -f20*s2/r
          ylms(ig,1,7)= 0.0d0
          ylms(ig,2,7)= 0.0d0
          ylms(ig,3,7)= f21*6.0d0*s3/r
          ylms(ig,1,8)= -f20*s3/r
          ylms(ig,2,8)= 0.0d0
          ylms(ig,3,8)= -f20*s1/r
          ylms(ig,1,9)= f22*2.0d0*s1/r
          ylms(ig,2,9)= -f22*2.0d0*s2/r
          ylms(ig,3,9)= 0.0d0
       end do
    else
       write(6,*)'LMAX must be >= 3.  LMAX=',lmax
       stop 'GETYL'
    end if
    return
  end subroutine getyls

  subroutine formod(for,dfor,bb,nsi)
    implicit none
    integer nsi
    double precision for(3,nsi),dfor(3,nsi),bb(3,3)
    double precision cc(3,3)
    double precision pai, dfpai
    parameter(pai = 3.14159265358979324d0)
    parameter(dfpai=1.0d0/(4.0d0*pai*pai))

    integer i,j,k,it

    do i=1,3
       do j=1,3
          cc(i,j)=0.0d0
          do k=1,3
             cc(i,j)=cc(i,j)+bb(k,i)*bb(k,j)*dfpai
          end do
       end do
    end do
    do it=1,nsi
       do i=1,3
          dfor(i,it)=cc(i,1)*for(1,it)+cc(i,2)*for(2,it)+cc(i,3)*for(3,it)
       end do
    end do
    return
  end subroutine formod

  !  VTRAN
  !   transform vector by TT(3,3)
  !     V(3,*) <-   TT(3,3)*V(3,*)
  !
  subroutine vtran(v,tt,n)
    implicit none
    integer:: n
    real(kind=8):: v(3,n),tt(3,3),a(3)
    integer:: i,j

    do i=1,n
       do j=1,3
          a(j)=tt(j,1)*v(1,i)+tt(j,2)*v(2,i)+tt(j,3)*v(3,i)
       end do
       do j=1,3
          v(j,i)=a(j)
       end do
    end do
    return
  end subroutine vtran

  subroutine cell2cart(aa,n,vcell,vcart)
    implicit none
    integer:: n
    real(kind=8):: aa(3,3),vcell(3,n),vcart(3,n)
    integer:: i,j,k

    vcart = 0.0d0
    do i=1,n
       do j=1,3
          do k=1,3
             vcart(k,i) = vcart(k,i) + aa(k,j)*vcell(j,i)
          end do
       end do
    end do
  end subroutine cell2cart

  subroutine cart2cell(aa,n,vcart,vcell)
    implicit none
    integer:: n
    real(kind=8):: aa(3,3),vcell(3,n),vcart(3,n)
    integer:: i,j,k
    real(kind=8):: aai(3,3),w

    w=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))

    aai(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/w
    aai(1,2)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/w
    aai(1,3)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/w
    aai(2,1)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/w
    aai(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/w
    aai(2,3)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/w
    aai(3,1)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/w
    aai(3,2)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/w
    aai(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/w

    vcell = 0.0d0
    do i=1,n
       do j=1,3
          do k=1,3
             vcell(k,i) = vcell(k,i) + aai(k,j)*vcart(j,i)
          end do
       end do
    end do
  end subroutine cart2cell

  subroutine cart2cellr(aa,n,vcart,vcellr)
    implicit none
    integer:: n
    real(kind=8):: aa(3,3),vcellr(3,n),vcart(3,n)
    integer:: i,j,k
    real(kind=8):: aai(3,3),w

    w=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))

    aai(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/w
    aai(1,2)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/w
    aai(1,3)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/w
    aai(2,1)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/w
    aai(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/w
    aai(2,3)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/w
    aai(3,1)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/w
    aai(3,2)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/w
    aai(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/w

    vcellr = 0.0d0
    do i=1,n
       do j=1,3
          do k=1,3
             vcellr(k,i) = vcellr(k,i) + aai(k,j)*vcart(j,i)
          end do
       end do
       do k=1,3
          vcellr(k,i) = modulo(vcellr(k,i),1.0d0)
       end do
    end do
  end subroutine cart2cellr

  !
  ! transform contravariant vectors
  ! from cartesian coordinate to cell coordinate
  !
  subroutine acart2cell(aa,n,vcart,vcell)
    implicit none
    integer:: n
    real(kind=8):: aa(3,3),vcart(3,n),vcell(3,n)
    integer:: i,j,k

    vcell = 0.0d0
    do i=1,n
       do j=1,3
          do k=1,3
             vcell(k,i) = vcell(k,i) + aa(j,k)*vcart(j,i)
          end do
       end do
    end do
  end subroutine acart2cell

  subroutine recp_uvec(aa,bb)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8),intent(out):: bb(3,3)
    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    real(kind=8):: v

    v=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))
    v=v/(2.0d0*pi)

    bb(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/v
    bb(2,1)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/v
    bb(3,1)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/v
    bb(1,2)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/v
    bb(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/v
    bb(3,2)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/v
    bb(1,3)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/v
    bb(2,3)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/v
    bb(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/v
    return
  end subroutine recp_uvec

  subroutine calchr(rho,nrxyz,va,chrg)
    implicit none
    integer,intent(in):: nrxyz
    real(kind=8),intent(in):: rho(nrxyz),va
    real(kind=8),intent(out):: chrg
    integer:: i
    chrg=0.0d0
    do i=1,nrxyz
       chrg=chrg+rho(i)
    end do
    chrg=chrg*va/nrxyz
    return
  end subroutine calchr

  subroutine hsort3i(n, kk, idx)
    implicit none
    integer,intent(in):: n
    integer:: kk(3,n)
    integer:: idx(n)

    integer:: i, j, ip, ic, it
    integer:: t(3)

    ! initialize idx
    do i=1,n
       idx(i) = i
    end do

    ! create the heap
    do i=2,n
       j = i
       ip = j/2
       do while ( gt3i(kk(1,j),kk(1,ip)) )
          t = kk(:,ip)
          it = idx(ip)
          kk(:,ip) = kk(:,j)
          idx(ip) = idx(j)
          kk(:,j) = t
          idx(j) = it
          j = ip
          ip = j/2
          if ( ip .eq. 0 ) exit
       end do
    end do

    ! sort it
    do i=n,1,-1
       t = kk(:,i)
       it = idx(i)
       kk(:,i) = kk(:,1)
       idx(i) = idx(1)
       kk(:,1) = t
       idx(1) = it
       j = 1
       ic = 2*j
       do while ( ic .lt. i )
          if ( ic + 1 .lt. i ) then
             if ( gt3i(kk(1,ic+1),kk(1,ic)) ) ic = ic + 1
          end if
          if ( gt3i(kk(1,ic),kk(1,j)) ) then
             t = kk(:,ic)
             it = idx(ic)
             kk(:,ic) = kk(:,j)
             idx(ic) = idx(j)
             kk(:,j) = t
             idx(j) = it
             j = ic
          else
             exit
          end if
          ic = 2*j
       end do
    end do
    return
    contains
      logical function gt3i(ka,kb)
        implicit none
        integer,intent(in):: ka(3),kb(3)

        if (ka(3).gt.kb(3)) then
           gt3i = .true.
           return
        else if (ka(3).eq.kb(3)) then
           if (ka(2).gt.kb(2)) then
              gt3i = .true.
              return
           else if (ka(2).eq.kb(2)) then
              if (ka(1).gt.kb(1)) then
                 gt3i = .true.
                 return
              end if
              gt3i = .false.
              return
           end if
           gt3i = .false.
           return
        end if
        gt3i = .false.
        return
      end function gt3i
  end subroutine hsort3i

  subroutine hanningwind(n,m,a,f,c1,c2)
    implicit none
    integer,intent(in):: n,m
    real(kind=8):: f(n,m)
    real(kind=8),intent(in):: a(n),c1,c2

    real(kind=8), parameter:: pi = 3.14159265358979323846d0
    integer:: i,j
    real(kind=8):: w

    do j=1,m
       do i=1,n
          if (a(i) .le. c1) then
             w = 1.0d0
          else if (a(i) .le. c2) then
             w = 0.5d0*( cos(pi/(c2-c1)*(a(i)-c1)) + 1.0d0 )
          else
             w = 0.0d0
          end if
          f(i,j) = w*f(i,j)
       end do
    end do
    return
  end subroutine hanningwind
  
  subroutine hsort(n, d, idx)
    implicit none
    integer,intent(in):: n
    real(kind=8):: d(n)
    integer:: idx(n)

    integer:: i, j, ip, ic, it
    real(kind=8):: t

    ! initialize idx
    do i=1,n
       idx(i) = i
    end do

    ! create the heap
    do i=2,n
       j = i
       ip = j/2
       do while ( d(j) .gt. d(ip) )
          t = d(ip)
          it = idx(ip)
          d(ip) = d(j)
          idx(ip) = idx(j)
          d(j) = t
          idx(j) = it
          j = ip
          ip = j/2
          if ( ip .eq. 0 ) exit
       end do
    end do

    ! sort it
    do i=n,1,-1
       t = d(i)
       it = idx(i)
       d(i) = d(1)
       idx(i) = idx(1)
       d(1) = t
       idx(1) = it
       j = 1
       ic = 2*j
       do while ( ic .lt. i )
          if ( ic + 1 .lt. i ) then
             if ( d(ic+1) .gt. d(ic) ) ic = ic + 1
          end if
          if ( d(ic) .gt. d(j) ) then
             t = d(ic)
             it = idx(ic)
             d(ic) = d(j)
             idx(ic) = idx(j)
             d(j) = t
             idx(j) = it
             j = ic
          else
             exit
          end if
          ic = 2*j
       end do
    end do
    return
  end subroutine hsort

end module util
